<?php

namespace Drupal\custom_elements\Entity;

use Drupal\Core\Entity\EntityDisplayBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldDefinition;
use Drupal\Core\Plugin\DefaultLazyPluginCollection as DefaultLazyPluginCollectionAlias;
use Drupal\custom_elements\CustomElementGenerator;

/**
 * Custom element display configuration entity.
 *
 * Contains custom element display options for all components of a rendered
 * entity in a given view mode.
 *
 * @ConfigEntityType(
 *   id = "entity_ce_display",
 *   label = @Translation("Entity custom element display"),
 *   entity_keys = {
 *     "id" = "id",
 *     "status" = "status"
 *   },
 *   handlers = {
 *     "access" = "\Drupal\custom_elements\Entity\Access\EntityCeDisplayAccessControlHandler",
 *   },
 *   config_export = {
 *     "id",
 *     "targetEntityType",
 *     "bundle",
 *     "mode",
 *     "customElementName",
 *     "forceAutoProcessing",
 *     "content",
 *     "hidden",
 *   }
 * )
 */
class EntityCeDisplay extends EntityDisplayBase implements EntityCeDisplayInterface {

  /**
   * {@inheritdoc}
   */
  protected $displayContext = 'view';

  /**
   * Whether this display is enabled or not.
   *
   * If the entity (form) display is disabled, we'll fall back to the 'default'
   * display.
   *
   * @var bool
   */
  protected $status = TRUE;

  /**
   * Custom element name to be displayed.
   *
   * @var string
   */
  protected $customElementName;

  /**
   * Status for 'ignore this config / fall back to using processors'.
   *
   * @var bool
   */
  protected $forceAutoProcessing = FALSE;

  /**
   * Custom elements generator.
   *
   * @var \Drupal\custom_elements\CustomElementGenerator
   */
  protected CustomElementGenerator $ceGenerator;

  /**
   * {@inheritdoc}
   */
  public function postCreate(EntityStorageInterface $storage) {

  }

  /**
   * {@inheritdoc}
   */
  public function __construct(array $values, $entity_type) {
    $this->pluginManager = \Drupal::service('custom_elements.plugin.manager.field.custom_element_formatter');
    $this->ceGenerator = \Drupal::service('custom_elements.generator');
    parent::__construct($values, $entity_type);
  }

  /**
   * {@inheritdoc}
   */
  public function setOriginalMode(string $viewMode): self {
    $this->originalMode = $viewMode;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getForceAutoProcessing(): bool {
    return !empty($this->forceAutoProcessing);
  }

  /**
   * {@inheritdoc}
   */
  public function setForceAutoProcessing(bool $status): self {
    $this->forceAutoProcessing = $status;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getCustomElementName() {
    return $this->customElementName;
  }

  /**
   * {@inheritdoc}
   */
  public function setCustomElementName($name) {
    $this->set('customElementName', $name);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getRenderer($field_name) {
    if (isset($this->plugins[$field_name])) {
      return $this->plugins[$field_name];
    }
    // Instantiate the formatter object from the stored display properties.
    if (($component = $this->getComponent($field_name)) && isset($component['formatter']) && ($definition = $this->getFieldDefinition($field_name))) {
      $component += ['configuration' => [], 'name' => $field_name, 'is_slot' => FALSE];
      $formatter = $this->pluginManager->createInstance($component['formatter'], [
        'field_definition' => $definition,
        'view_mode' => $this->originalMode,
        'name' => $component['name'],
        'is_slot' => $component['is_slot'],
      ] + $component['configuration']
      );
    }
    else {
      $formatter = NULL;
    }

    // Persist the formatter object.
    $this->plugins[$field_name] = $formatter;
    return $formatter;
  }

  /**
   * {@inheritdoc}
   */
  public function getPluginCollections() {
    $configurations = [];
    foreach ($this->getComponents() as $field_name => $component) {
      if (!empty($component['formatter']) && ($field_definition = $this->getFieldDefinition($field_name))) {
        $component += ['configuration' => [], 'name' => $field_name, 'is_slot' => FALSE];
        $configurations[$field_name] = [
          'id' => $component['formatter'],
          'field_definition' => $field_definition,
          'view_mode' => $this->originalMode,
          'name' => $component['name'],
          'is_slot' => $component['is_slot'],
        ] + $component['configuration'];
      }
    }
    return [
      'formatters' => new DefaultLazyPluginCollectionAlias($this->pluginManager, $configurations),
    ];
  }

  /**
   * {@inheritdoc}
   */
  protected function init() {
    // Do not call parent::init(); we don't need its display defaults.
    if (!$this->targetEntityType) {
      return;
    }

    if (!$this->getCustomElementName()) {
      $custom_element = $this->ceGenerator->getViewModeDefaults($this->targetEntityType, $this->bundle, $this->originalMode);
      $this->setCustomElementName($custom_element->getPrefixedTag());
    }

    // Check if any components are present before doing other initialization.
    $componentsPresent = !empty($this->content) || !empty($this->hidden);
    if (!$componentsPresent) {
      // Enable components options as done in the regular entity_view_display.
      $entity_view_display = \Drupal::service('entity_display.repository')
        ->getViewDisplay($this->targetEntityType, $this->bundle, $this->originalMode);
      $field_definitions = \Drupal::service('entity_field.manager')->getFieldDefinitions($this->targetEntityType, $this->bundle);

      // @todo Add support for statically set values.
      // Enable every component with "auto" that is enabled in the display.
      foreach ($entity_view_display->getComponents() as $name => $component) {
        // Ignore extra-fields.
        if (isset($field_definitions[$name])) {
          $default_name = str_starts_with($name, 'field_') ? substr($name, strlen('field_')) : $name;
          $this->setComponent($name, [
            'name' => $default_name,
            'formatter' => 'auto',
            'weight' => $component['weight'],
            'region' => 'content',
            'is_slot' => str_starts_with($field_definitions[$name]->getType(), 'text') ? 1 : 0,
          ]);
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function setComponent($name, array $options = []) {
    // If no weight specified, make sure the field sinks at the bottom.
    if (!isset($options['weight'])) {
      $max = $this->getHighestWeight();
      $options['weight'] = isset($max) ? $max + 1 : 0;
    }
    $this->content[$name] = $options;
    unset($this->hidden[$name]);
    unset($this->plugins[$name]);
    return $this;
  }

  /**
   * Gets the definitions of the fields that are candidate for display.
   */
  protected function getFieldDefinitions() {
    if (!isset($this->fieldDefinitions)) {
      $definitions = \Drupal::service('entity_field.manager')->getFieldDefinitions($this->targetEntityType, $this->bundle);

      $this->fieldDefinitions = $definitions;
      foreach ($this->fieldDefinitions as $field_definition) {
        if ($field_definition instanceof BaseFieldDefinition || $field_definition instanceof FieldDefinition) {
          // We'd like to be able to configure all the fields thus override
          // displayConfigurable to true for all field definitions.
          $field_definition->setDisplayConfigurable($this->displayContext, TRUE);
        }
      }
    }

    return $this->fieldDefinitions;
  }

}
